def all_solid(dict0, dict1, force_inside01=True):
    """
    Draw all frontiers
    """
    return 0

def all_invisible(dict0, dict1, force_inside01=True):
    """
    Draw none frontiers
    """
    return 1

